<?php

declare(strict_types=1);

/*
 * Chill is a software for social workers
 *
 * For the full copyright and license information, please view
 * the LICENSE file that was distributed with this source code.
 */

namespace Chill\TicketBundle\Serializer\Normalizer;

use Chill\MainBundle\Entity\User;
use Chill\MainBundle\Entity\UserGroup;
use Chill\PersonBundle\Entity\Person;
use Chill\TicketBundle\Entity\AddresseeHistory;
use Chill\TicketBundle\Entity\Comment;
use Chill\TicketBundle\Entity\MotiveHistory;
use Chill\TicketBundle\Entity\PersonHistory;
use Chill\TicketBundle\Entity\Ticket;
use Symfony\Component\Serializer\Exception\UnexpectedValueException;
use Symfony\Component\Serializer\Normalizer\NormalizerAwareInterface;
use Symfony\Component\Serializer\Normalizer\NormalizerAwareTrait;
use Symfony\Component\Serializer\Normalizer\NormalizerInterface;

final class TicketNormalizer implements NormalizerInterface, NormalizerAwareInterface
{
    use NormalizerAwareTrait;

    public function normalize($object, ?string $format = null, array $context = [])
    {
        if (!$object instanceof Ticket) {
            throw new UnexpectedValueException();
        }

        return [
            'type' => 'ticket_ticket',
            'id' => $object->getId(),
            'externalRef' => $object->getExternalRef(),
            'currentPersons' => $this->normalizer->normalize($object->getPersons(), $format, [
                'groups' => 'read',
            ]),
            'currentAddressees' => $this->normalizer->normalize($object->getCurrentAddressee(), $format, ['groups' => 'read']),
            'currentInputs' => $this->normalizer->normalize($object->getCurrentInputs(), $format, ['groups' => 'read']),
            'currentMotive' => $this->normalizer->normalize($object->getMotive(), $format, ['groups' => 'read']),
            'history' => array_values($this->serializeHistory($object, $format, ['groups' => 'read'])),
            'createdAt' => $this->normalizer->normalize($object->getCreatedAt(), $format, $context),
            'updatedAt' => $this->normalizer->normalize($object->getUpdatedAt(), $format, $context),
            'updatedBy' => $this->normalizer->normalize($object->getUpdatedBy(), $format, $context),
            'createdBy' => $this->normalizer->normalize($object->getCreatedBy(), $format, $context),
        ];
    }

    public function supportsNormalization($data, ?string $format = null)
    {
        return 'json' === $format && $data instanceof Ticket;
    }

    private function serializeHistory(Ticket $ticket, string $format, array $context): array
    {
        $events = [
            ...array_map(
                fn (MotiveHistory $motiveHistory) => [
                    'event_type' => 'set_motive',
                    'at' => $motiveHistory->getStartDate(),
                    'by' => $motiveHistory->getCreatedBy(),
                    'data' => $motiveHistory,
                ],
                $ticket->getMotiveHistories()->toArray()
            ),
            ...array_map(
                fn (PersonHistory $personHistory) =>  [
                    'event_type' => 'add_person',
                    'at' => $personHistory->getStartDate(),
                    'by' => $personHistory->getCreatedBy(),
                    'data' => $personHistory,
                ],
                $ticket->getPersonHistories()->toArray(),
            ),
            ...array_map(
                fn (Comment $comment) => [
                    'event_type' => 'add_comment',
                    'at' => $comment->getCreatedAt(),
                    'by' => $comment->getCreatedBy(),
                    'data' => $comment,
                ],
                $ticket->getComments()->toArray(),
            ),
            /*
            ...array_map(
                fn (AddresseeHistory $history) => [
                    'event_type' => 'add_addressee',
                    'at' => $history->getStartDate(),
                    'by' => $history->getCreatedBy(),
                    'data' => $history,
                ],
                $ticket->getAddresseeHistories()->toArray(),
            ),
            ...array_map(
                fn (AddresseeHistory $history) => [
                    'event_type' => 'remove_addressee',
                    'at' => $history->getStartDate(),
                    'by' => $history->getRemovedBy(),
                    'data' => $history,
                ],
                $ticket->getAddresseeHistories()->filter(fn (AddresseeHistory $history) => null !== $history->getEndDate())->toArray()
            ),
            */
            ...$this->addresseesStates($ticket),
            ...$this->personStates($ticket),
        ];

        if (null !== $ticket->getCreatedBy() && null !== $ticket->getCreatedAt()) {
            $events[] =
                [
                    'event_type' => 'create_ticket',
                    'at' => $ticket->getCreatedAt()->sub(new \DateInterval('PT1S')), // TODO hack to avoid collision with creation of the ticket event,
                    'by' => $ticket->getCreatedBy(),
                    'data' => [],
                ];
        }

        usort(
            $events,
            static function (array $a, array $b): int {
                return $a['at'] <=> $b['at'];
            }
        );

        return array_map(
            fn ($data) => [
                'event_type' => $data['event_type'],
                'at' => $this->normalizer->normalize($data['at'], $format, $context),
                'by' => $this->normalizer->normalize($data['by'], $format, $context),
                'data' => $this->normalizer->normalize($data['data'], $format, $context),
            ],
            $events
        );
    }

    private function addresseesStates(Ticket $ticket): array
    {
        /** @var array{string, array{added: list<AddresseeHistory>, removed: list<AddresseeHistory>}} $changes */
        $changes = [];
        $dateFormat = 'Y-m-d-m-Y-H-i-s';

        foreach ($ticket->getAddresseeHistories() as $history) {
            $changes[$history->getStartDate()->format($dateFormat)]['added'][] = $history;
            if (null !== $history->getEndDate()) {
                $changes[$history->getEndDate()->format($dateFormat)]['removed'][] = $history;
            }
        }

        ksort($changes);

        $currents = [];
        $steps = [];
        foreach ($changes as $change) {
            $historiesAdded = $change['added'] ?? [];
            $historiesRemoved = $change['removed'] ?? [];

            if (0 < count($historiesAdded)) {
                $at = $historiesAdded[0]->getStartDate();
                $by = $historiesAdded[0]->getCreatedBy();
            } elseif (0 < count($historiesRemoved)) {
                $at = $historiesRemoved[0]->getEndDate();
                $by = $historiesRemoved[0]->getRemovedBy();
            } else {
                throw new \LogicException('it should have at least one history');
            }

            $removed = array_map(fn (AddresseeHistory $history) => $history->getAddressee(), $historiesRemoved);
            $currents = array_filter($currents, fn (User|UserGroup $a) => !in_array($a, $removed, true));
            foreach ($historiesAdded as $history) {
                $currents[] = $history->getAddressee();
            }

            $steps[] = [
                'event_type' => 'addressees_state',
                'at' => $at,
                'by' => $by,
                'data' => [
                    'addressees' => array_values($currents),
                ],
            ];
        }

        return $steps;
    }

    private function personStates(Ticket $ticket): array
    {
        /** @var array{string, array{added: list<PersonHistory>, removed: list<PersonHistory>}} $changes */
        $changes = [];
        $dateFormat = 'Y-m-d-m-Y-H-i-s';

        foreach ($ticket->getPersonHistories() as $history) {
            $changes[$history->getStartDate()->format($dateFormat)]['added'][] = $history;
            if (null !== $history->getEndDate()) {
                $changes[$history->getEndDate()->format($dateFormat)]['removed'][] = $history;
            }
        }

        ksort($changes);

        $currents = [];
        $steps = [];
        foreach ($changes as $change) {
            $historiesAdded = $change['added'] ?? [];
            $historiesRemoved = $change['removed'] ?? [];

            if (0 < count($historiesAdded)) {
                $at = $historiesAdded[0]->getStartDate();
                $by = $historiesAdded[0]->getCreatedBy();
            } elseif (0 < count($historiesRemoved)) {
                $at = $historiesRemoved[0]->getEndDate();
                $by = $historiesRemoved[0]->getRemovedBy();
            } else {
                throw new \LogicException('it should have at least one history');
            }

            $removed = array_map(fn (PersonHistory $history) => $history->getPerson(), $historiesRemoved);
            $currents = array_filter($currents, fn (Person $a) => !in_array($a, $removed, true));
            foreach ($historiesAdded as $history) {
                $currents[] = $history->getPerson();
            }

            $steps[] = [
                'event_type' => 'persons_state',
                'at' => $at,
                'by' => $by,
                'data' => [
                    'persons' => array_values($currents),
                ],
            ];
        }

        return $steps;
    }
}
